"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""

class LudolphPlugin(object):
    """
    Ludolph plugin base class.
    """
    def __init__(self, config, *args, **kwargs):
        pass
