"""
Ludolph: Monitoring Jabber Bot
Version: 0.3.0
Homepage: https://github.com/erigones/Ludolph
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

__version__ = '0.3.0'
